
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <ctype.h>

#include <GL/gl.h>

#include "defines.h"
#include "atofont_core.h"



/* reset the modelview to no rotations and no scaling */
int _atofont_billboard_spherical_cheat_apply(float x, float y, float z) {

  GLfloat m[4*4];
  int i;


  glPushMatrix();

  /* go to the new position */
  glTranslatef(x, y, z);

  glGetFloatv(GL_MODELVIEW_MATRIX, m);

  /* undo rotations and scaling */
  i = 0;

  m[i++] = 1.0f;
  m[i++] = 0.0f;
  m[i++] = 0.0f;
  i++;
  m[i++] = 0.0f;
  m[i++] = 1.0f;
  m[i++] = 0.0f;
  i++;
  m[i++] = 0.0f;
  m[i++] = 0.0f;
  m[i] = 1.0f;

  /* set the modelview with no rotations and scaling */
  glLoadMatrixf(m);

  return SUCCEEDED;
}


float _atofont_get_string_width(struct atofont *font, unsigned char *s) {

  struct atocharacter *c;
  float f, fx;
  int n;


  if (font == NULL || s == NULL)
    return 0.0f;

  f = 0.0f;
  n = 0;
  fx = font->scale_x + font->spacing;

  if (font->type == ATOFONT_TYPE_FIXED) {
    while (s[n] != 0) {
      f += fx;
      n++;
    }
  }
  else {
    while (s[n] != 0) {
      c = &(font->characters[(int)s[n++]]);
      f += c->width*font->scale_x + font->spacing;
    }
  }

  f -= font->spacing;

  return f;
}


int atofont_set_spacing(struct atofont *font, float spacing) {

  if (font == NULL)
    return FAILED;

  font->spacing = spacing;

  return SUCCEEDED;
}


int atofont_set_scaling(struct atofont *font, float scale_x, float scale_y) {

  if (font == NULL)
    return FAILED;

  font->scale_x = scale_x;
  font->scale_y = scale_y;

  /* recreate the vertices */
  font->vertices[0*2 + 0] = 0.0f;
  font->vertices[0*2 + 1] = 0.0f;
  font->vertices[1*2 + 0] = scale_x;
  font->vertices[1*2 + 1] = 0.0f;
  font->vertices[2*2 + 0] = 0.0f;
  font->vertices[2*2 + 1] = scale_y;
  font->vertices[3*2 + 0] = scale_x;
  font->vertices[3*2 + 1] = scale_y;

  return SUCCEEDED;
}


int atofont_set_type(struct atofont *font, int type) {

  struct atocharacter *ac;
  int i, x, y, cx, cy, a, b, dx, lm, rm, s;


  if (font == NULL)
    return FAILED;

  if (!(type == ATOFONT_TYPE_FIXED || type == ATOFONT_TYPE_PROPORTIONAL))
    return FAILED;

  font->type = type;

  cx = font->texture_dx / font->character_dx;
  cy = font->texture_dy / font->character_dy;
  dx = font->texture_dx;

  /* rebuild the character information table */
  for (i = 0; i < 256; i++) {
    ac = &(font->characters[i]);
    x = i % cx;
    y = i / cx;
    x *= font->character_dx;
    y *= font->character_dy;
    ac->x = ((float)x)/font->texture_dx;
    ac->y = ((float)y)/font->texture_dy;

    if (type == ATOFONT_TYPE_FIXED) {
      ac->width = 1.0f;
      ac->left_margin = 0.0f;
    }
    else {
      /* find the left margin */
      for (s = 0, b = 0; b < font->character_dx && s == 0; b++) {
	for (a = 0; a < font->character_dy; a++) {
	  if (font->texture_data[(y+a)*dx*4 + (x+b)*4 + 3] != 0) {
	    s = 1;
	    break;
	  }
	}
      }
      lm = b-1;

      /* find the right margin */
      for (s = 0, b = font->character_dx-1; b >= 0 && s == 0; b--) {
	for (a = 0; a < font->character_dy; a++) {
	  if (font->texture_data[(y+a)*dx*4 + (x+b)*4 + 3] != 0) {
	    s = 1;
	    break;
	  }
	}
      }
      rm = b+1;

      if (lm >= rm) {
	lm = 0;
	rm = font->character_dx/2;
      }

      ac->width = ((float)rm-lm+1)/((float)font->character_dx);
      ac->left_margin = ((float)lm)/((float)font->character_dx);
    }
  }

  return SUCCEEDED;
}


int atofont_free(struct atofont *font) {

  if (font == NULL)
    return FAILED;

  free(font);

  return SUCCEEDED;
}


int atofont_create(struct atofont **font, unsigned char *t, int type, int tx, int ty, int cx, int cy) {

  struct atofont *f;


  if (font == NULL || t == NULL)
    return FAILED;

  *font = NULL;

  if (type != GL_RGBA) {
    fprintf(stderr, "ATOFONT_CREATE: Only GL_RGBA-textures are supported.\n");
    return FAILED;
  }

  f = malloc(sizeof(struct atofont));
  if (f == NULL) {
    fprintf(stderr, "ATOFONT_CREATE: Out of memory error.\n");
    return FAILED;
  }

  f->texture_data = t;
  f->texture_type = type;
  f->texture_dx = tx;
  f->texture_dy = ty;
  f->character_dx = cx;
  f->character_dy = cy;
  f->character_fixed_width = ((float)cx)/((float)tx);
  f->character_fixed_height = ((float)cy)/((float)ty);

  atofont_set_spacing(f, 0.0f);
  atofont_set_scaling(f, 1.0f, 1.0f);
  atofont_set_type(f, ATOFONT_TYPE_FIXED);

  *font = f;

  return SUCCEEDED;
}
