
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <ctype.h>

#include <GL/gl.h>

#include "defines.h"
#include "atofont_core.h"
#include "atofont_so.h"



int atofont_string_object_free(struct atostringobject *so) {

  if (so == NULL)
    return FAILED;

  /* texturecoords and vertices are in the same memory block */
  if (so->texturecoords != NULL)
    free(so->texturecoords);

  free(so);

  return SUCCEEDED;
}


static void _atofont_string_object_write(struct atostringobject *so) {

  glVertexPointer(2, GL_FLOAT, 0, so->vertices);
  glTexCoordPointer(2, GL_FLOAT, 0, so->texturecoords);
  glDrawArrays(GL_TRIANGLES, 0, so->nvertices);
}


int atofont_string_object_create(struct atofont *font, char *s, struct atostringobject **so) {

  struct atostringobject *as;
  float *tc, *ve;
  unsigned int l;


  if (font == NULL || s == NULL || so == NULL)
    return FAILED;

  *so = NULL;

  /* count the characters */
  l = 0;
  while (s[l] != 0)
    l++;

  as = malloc(sizeof(struct atostringobject));
  tc = malloc(sizeof(float) * l*6*2 * 2);
  ve = &tc[l*6*2];
  if (as == NULL || tc == NULL) {
    fprintf(stderr, "ATOFONT_STRING_OBJECT_CREATE: Out of memory error.\n");
    return FAILED;
  }

  as->nvertices = l*2*3;
  as->width = _atofont_get_string_width(font, (unsigned char *)s);
  as->height = font->scale_y;
  as->vertices = ve;
  as->texturecoords = tc;

  atofont_string_object_recreate(font, s, as);

  *so = as;

  return SUCCEEDED;
}


int atofont_string_object_recreate(struct atofont *font, char *s, struct atostringobject *so) {

  struct atocharacter *c;
  float *tc, *ve;
  unsigned int i, k, n;
  float fg, m;


  if (font == NULL || s == NULL || so == NULL)
    return FAILED;

  i = 0;
  n = 0;
  m = 0.0f;
  fg = font->scale_x + font->spacing;
  tc = so->texturecoords;
  ve = so->vertices;

  while (s[i] != 0) {
    k = s[i++];
    c = &(font->characters[k]);

    if (font->type == ATOFONT_TYPE_PROPORTIONAL) {
      fg = c->left_margin*font->scale_x;
      m -= fg;
    }

    /* copy the vertices and texcoords */
    ve[n] = font->vertices[0*2 + 0] + m;
    tc[n] = c->x;
    n++;
    ve[n] = font->vertices[0*2 + 1];
    tc[n] = c->y + font->character_fixed_height;
    n++;

    ve[n] = font->vertices[1*2 + 0] + m;
    tc[n] = c->x + font->character_fixed_width;
    n++;
    ve[n] = font->vertices[1*2 + 1];
    tc[n] = c->y + font->character_fixed_height;
    n++;

    ve[n] = font->vertices[2*2 + 0] + m;
    tc[n] = c->x;
    n++;
    ve[n] = font->vertices[2*2 + 1];
    tc[n] = c->y;
    n++;

    ve[n] = font->vertices[1*2 + 0] + m;
    tc[n] = c->x + font->character_fixed_width;
    n++;
    ve[n] = font->vertices[1*2 + 1];
    tc[n] = c->y + font->character_fixed_height;
    n++;

    ve[n] = font->vertices[2*2 + 0] + m;
    tc[n] = c->x;
    n++;
    ve[n] = font->vertices[2*2 + 1];
    tc[n] = c->y;
    n++;

    ve[n] = font->vertices[3*2 + 0] + m;
    tc[n] = c->x + font->character_fixed_width;
    n++;
    ve[n] = font->vertices[3*2 + 1];
    tc[n] = c->y;
    n++;

    if (font->type == ATOFONT_TYPE_FIXED)
      m += fg;
    else
      m += fg + c->width*font->scale_x + font->spacing;
  }

  return SUCCEEDED;
}


int atofont_string_object_write_billboard(struct atostringobject *so, float x, float y, float z, int flags) {

  if (so == NULL)
    return FAILED;

  /* fix the modelview */
  _atofont_billboard_spherical_cheat_apply(x, y, z);

  if ((flags & ATOFONT_CENTERED) != 0)
    glTranslatef(-so->width/2.0f, -so->height/2.0f, 0);

  /* write the string */
  _atofont_string_object_write(so);

  /* return the modelview */
  glPopMatrix();

  return SUCCEEDED;
}


int atofont_string_object_write_2D(struct atostringobject *so, float x, float y, int flags) {

  if (so == NULL)
    return FAILED;

  glPushMatrix();

  if ((flags & ATOFONT_CENTERED) != 0)
    glTranslatef(-so->width/2.0f + x, -so->height/2.0f + y, 0);
  else
    glTranslatef(x, y, 0);

  /* write the string */
  _atofont_string_object_write(so);

  glPopMatrix();

  return SUCCEEDED;
}
